import { DependencyList, useEffect, useRef } from 'react'

type Args = Parameters<typeof window.history.pushState>
type PopEvent = PopStateEvent & { args?: Args }
type PushAndReplaceEvent = Event & { args?: Args }
type UnionEvent = PopEvent | PushAndReplaceEvent
interface IVisitorCallbacks {
  // 后退
  onPopState?: (data?: Args[0]) => void
  onPushState?: (data?: Args[0]) => void
  onReplaceState?: (data?: Args[0]) => void
}

const _wr = (type: 'pushState' | 'replaceState') => {
  const origin = window.history[type]
  return function (...args: Args) {
    // @ts-ignore
    // eslint-disable-next-line @typescript-eslint/no-invalid-this
    const newEvent = origin.apply(this, args)
    const e = new Event(type)
    ;(e as any).args = args
    window.dispatchEvent(e)
    return newEvent
  }
}

export function useVisitor(callbacks: IVisitorCallbacks, deps: DependencyList = []) {
  const timer = useRef<ReturnType<typeof setTimeout>>()
  const callbacksRef = useRef<IVisitorCallbacks>(callbacks)

  const onStateChange = (type: keyof IVisitorCallbacks, e: UnionEvent) => {
    if (timer.current) {
      clearTimeout(timer.current)
    }
    timer.current = setTimeout(() => {
      let data = (e as PopStateEvent).state
      if (e.args?.[0]) {
        data = e.args[0]
      }
      callbacksRef.current[type]?.(data)
    }, 5)
  }

  const onPopState = (e: PopStateEvent) => {
    onStateChange('onPopState', e as UnionEvent)
  }
  const onPushState = (e: Event) => {
    onStateChange('onPushState', e as UnionEvent)
  }

  const onReplaceState = (e: Event) => {
    onStateChange('onReplaceState', e as UnionEvent)
  }

  useEffect(() => {
    callbacksRef.current = callbacks
  }, [deps, callbacks])

  const unregister = () => {
    window.removeEventListener('popstate', onPopState)
    window.removeEventListener('pushState', onPushState)
    window.removeEventListener('replaceState', onReplaceState)
  }

  const register = () => {
    window.addEventListener('popstate', onPopState)
    window.addEventListener('pushState', onPushState)
    window.addEventListener('replaceState', onReplaceState)
  }

  useEffect(() => {
    window.history.pushState = _wr('pushState')
    window.history.replaceState = _wr('replaceState')

    register()
    return () => {
      unregister()
      clearTimeout(timer.current)
    }
  }, [])
}
